# -*- coding: utf-8 -*-
"""
lists
~~~~~

The `lists` module contains standard collections based on Python lists.
Included collections are :class:`List` and :class:`Deque`.
Each collection stores its values in a Redis
`list <http://redis.io/commands#list>`_ structure.

.. note::
    If you need to store mutable values like :class:`list`\s or :class:`set`\s
    in a collection, be sure to enable ``writeback``.
    See :ref:`Synchronization` for more information.
"""
from __future__ import division, print_function, unicode_literals

import collections
import itertools
import uuid

import six
from redis import ResponseError

from .base import RedisCollection


class List(RedisCollection, collections.MutableSequence):
    """
    Collection based on the built-in Python :class:`list` type.
    Items are stored in a Redis list structure.
    See Python's `list documentation
    <https://docs.python.org/3/library/stdtypes.html#list>`_ for usage notes.
    """
    _python_cls = list

    def __init__(self, *args, **kwargs):
        """
        Create a new List object.

        If the first argument (*data*) is an iterable object, create the new
        List with its values as the initial data.

        :param data: Initial data.
        :type data: iterable
        :param redis: Redis client instance. If not provided, default Redis
                      connection is used.
        :type redis: :class:`redis.StrictRedis`
        :param key: Redis key for the collection. Collections with the same key
                    point to the same data. If not provided, a random
                    string is generated.
        :type key: str
        :param writeback: If ``True`` keep a local cache of changes for storing
                          modifications to mutable values. Changes will be
                          written to Redis after calling the ``sync`` method.
        :type writeback: bool
        """
        data = args[0] if args else kwargs.pop('data', None)
        writeback = kwargs.pop('writeback', False)
        super(List, self).__init__(**kwargs)

        self.__marker = uuid.uuid4().hex
        self.writeback = writeback
        self.cache = {}

        if data:
            self.extend(data)

    def _pop_left(self):
        # Retrieve a value from the 0 index, remove it, and return it.
        pickled_value = self.redis.lpop(self.key)
        if pickled_value is None:
            raise IndexError
        value = self._unpickle(pickled_value)

        if self.writeback:
            value = self.cache.get(0, value)
            items = six.iteritems(self.cache)
            self.cache = {i - 1: v for i, v in items if i != 0}

        return value

    def _pop_right(self):
        # Retrieve a value from the -1 index, remove it, and return it.
        if not self.writeback:
            pickled_value = self.redis.rpop(self.key)
            if pickled_value is None:
                raise IndexError
            return self._unpickle(pickled_value)

        # If writeback is enabled we'll need the size of the list; compute that
        # in a transaction
        def pop_right_trans(pipe):
            len_self, cache_index = self._normalize_index(-1, pipe=pipe)
            if len_self == 0:
                raise IndexError
            pickled_value = pipe.rpop(self.key)
            value = self.cache.get(cache_index, self._unpickle(pickled_value))
            items = six.iteritems(self.cache)
            self.cache = {i: v for i, v in items if i != cache_index}

            return value

        return self._transaction(pop_right_trans)

    def _pop_middle(self, index):
        # Retrieve the value at *index*, remove it, and return it.
        def pop_middle_trans(pipe):
            len_self, cache_index = self._normalize_index(index, pipe)
            if (cache_index < 0) or (cache_index >= len_self):
                raise IndexError

            # Retrieve the value at index, then overwrite it with a special
            # marker, and then remove the marker.
            value = self._unpickle(pipe.lindex(self.key, index))
            pipe.multi()
            pipe.lset(self.key, index, self.__marker)
            pipe.lrem(self.key, 1, self.__marker)

            if self.writeback:
                value = self.cache.get(cache_index, value)
                new_cache = {}
                for i, v in six.iteritems(self.cache):
                    if i < cache_index:
                        new_cache[i] = v
                    elif i > cache_index:
                        new_cache[i - 1] = v
                self.cache = new_cache

            return value

        return self._transaction(pop_middle_trans)

    def _del_slice(self, index):
        # Delete the values associated with the slice object *index*
        def del_slice_trans(pipe):
            start, stop, step, forward, len_self = self._normalize_slice(
                index, pipe
            )

            # Empty slice: nothing to do
            if start == stop:
                return

            # Write back the cache before making changes
            if self.writeback:
                self._sync_helper(pipe)

            # Steps must be done index by index
            if index.step is not None:
                pipe.multi()
                for i in list(six.moves.xrange(len_self))[index]:
                    pipe.lset(self.key, i, self.__marker)
                pipe.lrem(self.key, 0, self.__marker)
            # Slice covers entire range: delete the whole list
            elif start == 0 and stop == len_self:
                self.clear(pipe)
            # Slice starts on the left: keep the right
            elif start == 0 and stop != len_self:
                pipe.ltrim(self.key, stop, -1)
            # Slice stops on the right: keep the left
            elif start != 0 and stop == len_self:
                pipe.ltrim(self.key, 0, start - 1)
            # Slice starts and ends in the middle
            else:
                left_values = pipe.lrange(self.key, 0, start - 1)
                right_values = pipe.lrange(self.key, stop, -1)
                pipe.delete(self.key)
                all_values = itertools.chain(left_values, right_values)
                pipe.rpush(self.key, *all_values)

        return self._transaction(del_slice_trans)

    def __delitem__(self, index):
        """Delete the item at *index*."""
        if isinstance(index, slice):
            return self._del_slice(index)

        if index == 0:
            self._pop_left()
        elif index == -1:
            self._pop_right()
        else:
            self._pop_middle(index)

    def _get_slice(self, index):
        # Return the values specified by the slice object *index* as a Python
        # list

        def get_slice_trans(pipe):
            start, stop, step, forward, len_self = self._normalize_slice(
                index, pipe
            )

            if start == stop:
                return []

            ret = []
            redis_values = pipe.lrange(self.key, start, max(stop - 1, 0))
            for i, v in enumerate(redis_values, start):
                ret.append(self.cache.get(i, self._unpickle(v)))

            if not forward:
                ret = reversed(ret)

            if step != 1:
                ret = itertools.islice(ret, None, None, step)

            return list(ret)

        return self._transaction(get_slice_trans)

    def __getitem__(self, index):
        """
        If *index* is an :obj:`int`, return the value at that index.
        If *index* is a :obj:`slice`, return the values from that slice
        as a :obj:`list`.
        """
        if isinstance(index, slice):
            return self._get_slice(index)

        # If writeback is off we can just query Redis, since its indexing
        # scheme matches Python's. If the index is out of range we get None.
        if not self.writeback:
            pickled_value = self.redis.lindex(self.key, index)
            if pickled_value is None:
                raise IndexError

            return self._unpickle(pickled_value)

        # If writeback is on we'll need to know the size of the list,
        # so we'll need to use a transaction
        def getitem_trans(pipe):
            len_self, cache_index = self._normalize_index(index, pipe)

            if (cache_index < 0) or (cache_index >= len_self):
                raise IndexError

            if cache_index in self.cache:
                return self.cache[cache_index]

            value = self._unpickle(pipe.lindex(self.key, index))
            self.cache[cache_index] = value
            return value

        return self._transaction(getitem_trans)

    def _data(self, pipe=None):
        """
        Return a :obj:`list` of all values from Redis
        (without checking the local cache).
        """
        pipe = self.redis if pipe is None else pipe
        return [self._unpickle(v) for v in pipe.lrange(self.key, 0, -1)]

    def __iter__(self, pipe=None):
        """
        Return a :obj:`list` of all values from Redis (overriding those with
        values from the local cache)
        """
        return (self.cache.get(i, v) for i, v in enumerate(self._data(pipe)))

    def __len__(self, pipe=None):
        """Return the length of this collection."""
        pipe = self.redis if pipe is None else pipe
        return pipe.llen(self.key)

    def __reversed__(self):
        """
        Return an iterator over this collection's items in reverse order.
        """
        return reversed(list(self.__iter__()))

    def _set_slice(self, index, value):
        # Set the values for the indexes associated with the slice object
        # *index* to the contents of the iterable *value*
        def set_slice_trans(pipe):
            start, stop, step, forward, len_self = self._normalize_slice(
                index, pipe
            )

            # Write back the cache before making changes
            if self.writeback:
                self._sync_helper(pipe)

            # Loop through each index for slices with steps
            if index.step is not None:
                new_values = list(value)
                change_indexes = six.moves.xrange(start, stop, step)
                if len(new_values) != len(change_indexes):
                    raise ValueError
                for i, v in six.moves.zip(change_indexes, new_values):
                    pipe.lset(self.key, i, self._pickle(v))
            # For slices without steps retrieve the items to the left and right
            # of the slice, clear the collection, then re-insert the items
            # with the new values in the middle.
            else:
                if start == 0:
                    left_values = []
                else:
                    left_values = pipe.lrange(self.key, 0, start - 1)

                middle_values = (self._pickle(v) for v in value)

                if stop == len_self:
                    right_values = []
                else:
                    right_values = pipe.lrange(self.key, stop, -1)

                pipe.delete(self.key)
                all_values = itertools.chain(
                    left_values, middle_values, right_values
                )
                pipe.rpush(self.key, *all_values)

        return self._transaction(set_slice_trans)

    def __setitem__(self, index, value):
        """
        If *index* is an :obj:`int`, set the value for that index to *value*.
        If *index* is a :obj:`slice`, set the values for the indexes
        associated with that slice  to the contents of the iterable *value*.
        """
        if isinstance(index, slice):
            return self._set_slice(index, value)

        def setitem_trans(pipe):
            if self.writeback:
                __, cache_index = self._normalize_index(index, pipe=pipe)

            try:
                pipe.lset(self.key, index, self._pickle(value))
            except ResponseError:
                raise IndexError

            if self.writeback:
                self.cache[cache_index] = value

        self._transaction(setitem_trans)

    def append(self, value):
        """Insert *value* at the end of this collection."""
        len_self = self.redis.rpush(self.key, self._pickle(value))

        if self.writeback:
            self.cache[len_self - 1] = value

    def clear(self, pipe=None):
        """Delete all values from this collection."""
        self._clear(pipe)

        if self.writeback:
            self.cache.clear()

    def copy(self, key=None):
        """
        Return a new collection with the same items as this one.
        If *key* is specified, create the new collection with the given
        Redis key.
        """
        other = self.__class__(
            redis=self.redis, key=key, writeback=self.writeback
        )
        other.extend(self)

        return other

    def count(self, value):
        """
        Return the number of occurences of *value*.

        .. note::
            Counting is implemented in Python.
        """
        return sum(1 for v in self.__iter__() if v == value)

    def extend(self, other):
        """
        Adds the values from the iterable *other* to the end of this
        collection.
        """
        def extend_trans(pipe):
            values = list(other.__iter__(pipe)) if use_redis else other
            len_self = pipe.rpush(self.key, *(self._pickle(v) for v in values))
            if self.writeback:
                for i, v in enumerate(values, len_self - len(values)):
                    self.cache[i] = v

        if self._same_redis(other, RedisCollection):
            use_redis = True
            self._transaction(extend_trans, other.key)
        else:
            use_redis = False
            self._transaction(extend_trans)

    def index(self, value, start=None, stop=None):
        """
        Return the index of the first occurence of *value*.
        If *start* or *stop* are provided, return the smallest
        index such that ``s[index] == value`` and ``start <= index < stop``.
        """
        def index_trans(pipe):
            len_self, normal_start = self._normalize_index(start or 0, pipe)
            __, normal_stop = self._normalize_index(stop or len_self, pipe)
            for i, v in enumerate(self.__iter__(pipe=pipe)):
                if v == value:
                    if i < normal_start:
                        continue
                    if i >= normal_stop:
                        break
                    return i
            raise ValueError

        return self._transaction(index_trans)

    def _insert_left(self, value, pipe=None):
        # Insert *value* at index 0.
        pipe = self.redis if pipe is None else pipe
        pipe.lpush(self.key, self._pickle(value))
        if self.writeback:
            self.cache = {k + 1: v for k, v in six.iteritems(self.cache)}
            self.cache[0] = value

    def _insert_middle(self, index, value, pipe=None):
        # Insert *value* at *index*.
        pipe = self.redis if pipe is None else pipe

        # First, retrieve everything from the index to the end.
        __, cache_index = self._normalize_index(index, pipe)
        right_values = pipe.lrange(self.key, cache_index, -1)

        # Next, zap everything after the index. Finally, insert the new value
        # and then re-insert the items from before.
        pipe.multi()
        pipe.ltrim(self.key, 0, cache_index - 1)
        pipe.rpush(self.key, self._pickle(value), *right_values)

        if self.writeback:
            new_cache = {}
            for i, v in six.iteritems(self.cache):
                if i < cache_index:
                    new_cache[i] = v
                elif i >= cache_index:
                    new_cache[i + 1] = v
            new_cache[cache_index] = value
            self.cache = new_cache

    def insert(self, index, value):
        """
        Insert *value* into the collection at *index*.
        """
        if index == 0:
            return self._insert_left(value)

        def insert_middle_trans(pipe):
            self._insert_middle(index, value, pipe=pipe)

        return self._transaction(insert_middle_trans)

    def pop(self, index=-1):
        """
        Retrieve the value at *index*, remove it from the collection, and
        return it.
        """
        if index == 0:
            return self._pop_left()
        elif index == -1:
            return self._pop_right()
        else:
            return self._pop_middle(index)

    def remove(self, value):
        """Remove the first occurence of *value*."""
        def remove_trans(pipe):
            # If we're caching, we'll need to synchronize before removing.
            if self.writeback:
                self._sync_helper(pipe)

            delete_count = pipe.lrem(self.key, 1, self._pickle(value))
            if delete_count == 0:
                raise ValueError

        self._transaction(remove_trans)

    def reverse(self):
        """
        Reverses the items of this collection "in place" (only two values are
        retrieved from Redis at a time).
        """
        def reverse_trans(pipe):
            if self.writeback:
                self._sync_helper(pipe)

            n = self.__len__(pipe)
            for i in six.moves.xrange(n // 2):
                left = pipe.lindex(self.key, i)
                right = pipe.lindex(self.key, n - i - 1)
                pipe.lset(self.key, i, right)
                pipe.lset(self.key, n - i - 1, left)

        self._transaction(reverse_trans)

    def sort(self, key=None, reverse=False):
        """
        Sort the items of this collection according to the optional callable
        *key*. If *reverse* is set then the sort order is reversed.

        .. note::
            This sort requires all items to be retrieved from Redis and stored
            in memory.
         """
        def sort_trans(pipe):
            values = list(self.__iter__(pipe))
            values.sort(key=key, reverse=reverse)

            pipe.multi()
            pipe.delete(self.key)
            pipe.rpush(self.key, *(self._pickle(v) for v in values))

            if self.writeback:
                self.cache = {}

        return self._transaction(sort_trans)

    def _add_helper(
        self, other, use_redis=False, swap_args=False, **kwargs
    ):
        def add_helper_trans(pipe):
            self_values = self._python_cls(self.__iter__(pipe), **kwargs)
            if use_redis:
                other_values = self._python_cls(other.__iter__(pipe), **kwargs)
            else:
                other_values = self._python_cls(other, **kwargs)

            if swap_args:
                return other_values + self_values

            return self_values + other_values

        watches = [other.key] if use_redis else []
        return self._transaction(add_helper_trans, *watches)

    def __add__(self, other):
        if self._same_redis(other, RedisCollection):
            return self._add_helper(other, use_redis=True)

        return self._add_helper(other)

    def __radd__(self, other):
        return self._add_helper(other, swap_args=True)

    def __iadd__(self, other):
        self.extend(other)
        return self

    def __eq__(self, other):
        if not isinstance(other, (self.__class__, self._python_cls)):
            return False

        def eq_trans(pipe):
            self_values = self.__iter__(pipe)
            self_len = self.__len__(pipe)

            if use_redis:
                other_values = other.__iter__(pipe)
                other_len = other.__len__(pipe)
            else:
                other_values = other
                other_len = len(other)

            if self_len != other_len:
                return False

            for v_self, v_other in six.moves.zip(self_values, other_values):
                if v_self != v_other:
                    return False

            return True

        if self._same_redis(other, RedisCollection):
            use_redis = True
            return self._transaction(eq_trans, other.key)
        else:
            use_redis = False
            return self._transaction(eq_trans)

    def __mul__(self, times):
        if not isinstance(times, six.integer_types):
            raise TypeError

        return self._python_cls(self.__iter__()) * times

    def __rmul__(self, times):
        return self.__mul__(times)

    def __imul__(self, times):
        if not isinstance(times, six.integer_types):
            raise TypeError

        # If multiplying by 1 there's no work to do
        if times == 1:
            return self

        def imul_trans(pipe):
            # If multiplying by 0 or a negative number all values are deleted
            if times <= 0:
                self.clear(pipe)

            # Synchronize the cache before writing
            if self.writeback:
                self._sync_helper(pipe)

            # Pull in pickled values
            pickled_values = pipe.lrange(self.key, 0, -1)
            pipe.multi()

            # Write the values repeatedly
            for __ in six.moves.xrange(times - 1):
                pipe.rpush(self.key, *pickled_values)

        self._transaction(imul_trans)
        return self

    def _repr_data(self):
        items = (repr(v) for v in self.__iter__())
        return '[{}]'.format(', '.join(items))

    def _sync_helper(self, pipe):
        for i, v in six.iteritems(self.cache):
            pipe.lset(self.key, i, self._pickle(v))

        self.cache = {}

    def sync(self):
        def sync_trans(pipe):
            pipe.multi()
            self._sync_helper(pipe)

        self._transaction(sync_trans)


class Deque(List):
    """
    Collection based on the Python standard library's
    :class:`collections.deque` type.
    Items are stored in a Redis hash structure.
    See Python's `deque documentation
    <https://docs.python.org/3/library/collections.html#collections.deque>`_
    for usage notes.

    Dequq inherits from List, so see its API documentation for
    information on other methods.
    """
    _python_cls = collections.deque

    def __init__(self, iterable=None, maxlen=None, **kwargs):
        """
        Create a new Deque object.

        If the first argument (*iterable*) is an iterable object, create the
        new Deque with its values as the initial data.

        If the second argument (*maxlen*) is an integer, create the Deque with
        the given maximum length.
        If the second argument is not given or is ``None``, create the Deque
        without a maximum length.

        If the Deque is full (the number of values stored is equal to the
        maximum length), adding new items to one side will cause a
        corresponding number of items to be removed from the other side.

        :param iterable: Initial data.
        :type data: iterable
        :param maxlen: Maximum size.
        :type maxlen: int
        :param redis: Redis client instance. If not provided, default Redis
                      connection is used.
        :type redis: :class:`redis.StrictRedis`
        :param key: Redis key for the collection. Collections with the same key
                    point to the same data. If not provided, a random
                    string is generated.
        :type key: str
        :param writeback: If ``True`` keep a local cache of changes for storing
                          modifications to mutable values. Changes will be
                          written to Redis after calling the ``sync`` method.
        :type key: bool

        .. warning::
            The ``maxlen`` of the collection can't be enforced when multiple
            processes are accessing its Redis collection.
        """
        if iterable is not None:
            kwargs['data'] = iterable

        if (maxlen is not None) and not isinstance(maxlen, six.integer_types):
            raise TypeError('an integer is required')

        if (maxlen is not None) and maxlen < 0:
            raise ValueError('maxlen must be non-negative')

        self.maxlen = maxlen
        super(Deque, self).__init__(**kwargs)

    # Magic methods

    def __delitem__(self, index):
        if isinstance(index, slice):
            raise TypeError

        return super(Deque, self).__delitem__(index)

    def __getitem__(self, index):
        if isinstance(index, slice):
            raise TypeError

        return super(Deque, self).__getitem__(index)

    def __setitem__(self, index, value):
        if isinstance(index, slice):
            raise TypeError

        return super(Deque, self).__setitem__(index, value)

    # Named methods

    def _append_helper(self, value, pipe):
        # Append on the right
        len_self = pipe.rpush(self.key, self._pickle(value))
        if self.writeback:
            self.cache[len_self - 1] = value

        # Check the length restriction
        if (self.maxlen is None) or (len_self <= self.maxlen):
            return

        # Pop from the left
        pipe.lpop(self.key)
        if self.writeback:
            items = six.iteritems(self.cache)
            self.cache = {i - 1: v for i, v in items if i != 0}

    def append(self, value):
        """Add *value* to the right side of the collection."""
        def append_trans(pipe):
            self._append_helper(value, pipe)

        self._transaction(append_trans)

    def _appendleft_helper(self, value, pipe):
        # Append on the left
        len_self = pipe.lpush(self.key, self._pickle(value))
        if self.writeback:
            self.cache = {k + 1: v for k, v in six.iteritems(self.cache)}
            self.cache[0] = value

        # Check the length restriction
        if (self.maxlen is None) or (len_self <= self.maxlen):
            return

        # Pop from the right
        pipe.rpop(self.key)
        if self.writeback:
            cache_index = len_self - 1
            items = six.iteritems(self.cache)
            self.cache = {i: v for i, v in items if i != cache_index}

    def appendleft(self, value):
        """Add *value* to the left side of the collection."""
        def appendleft_trans(pipe):
            self._appendleft_helper(value, pipe)

        self._transaction(appendleft_trans)

    def copy(self, key=None):
        """
        Return a new collection with the same items as this one.
        If *key* is specified, create the new collection with the given
        Redis key.
        """
        other = self.__class__(
            self.__iter__(),
            self.maxlen,
            redis=self.redis,
            key=key,
            writeback=self.writeback,
        )

        return other

    def extend(self, other):
        """
        Extend the right side of the the collection by appending values from
        the iterable *other*.
        """
        def extend_trans(pipe):
            values = list(other.__iter__(pipe)) if use_redis else other
            for v in values:
                self._append_helper(v, pipe)

        if self._same_redis(other, RedisCollection):
            use_redis = True
            self._transaction(extend_trans, other.key)
        else:
            use_redis = False
            self._transaction(extend_trans)

    def extendleft(self, other):
        """
        Extend the left side of the the collection by appending values from
        the iterable *other*. Note that the appends will reverse the order
        of the given values.
        """
        def extendleft_trans(pipe):
            values = list(other.__iter__(pipe)) if use_redis else other
            for v in values:
                self._appendleft_helper(v, pipe)

        if self._same_redis(other, RedisCollection):
            use_redis = True
            self._transaction(extendleft_trans, other.key)
        else:
            use_redis = False
            self._transaction(extendleft_trans)

    def insert(self, index, value):
        """
        Insert *value* into the collection at *index*.
        If the insertion would the collection to grow beyond ``maxlen``,
        raise ``IndexError``.
        """
        def insert_trans(pipe):
            len_self = self.__len__(pipe)
            if (self.maxlen is not None) and (len_self >= self.maxlen):
                raise IndexError

            if index == 0:
                self._insert_left(value, pipe)
            else:
                self._insert_middle(index, value, pipe=pipe)

        self._transaction(insert_trans)

    def pop(self):
        """
        Remove and return an value from the right side of the collection.
        """
        return self._pop_right()

    def popleft(self):
        """
        Remove and return an value from the right side of the collection.
        """
        return self._pop_left()

    def rotate(self, n=1):
        """
        Rotate the deque n steps to the right.
        If n is negative, rotate to the left.
        """
        # No work to do for a 0-step rotate
        if n == 0:
            return

        def rotate_trans(pipe):
            # Synchronize the cache before rotating
            if self.writeback:
                self._sync_helper(pipe)

            # Rotating len(self) times has no effect.
            len_self = self.__len__(pipe)
            steps = abs_n % len_self

            # When n is positive we can use the built-in Redis command
            if forward:
                pipe.multi()
                for __ in six.moves.xrange(steps):
                    pipe.rpoplpush(self.key, self.key)
            # When n is negative we must use Python
            else:
                for __ in six.moves.xrange(steps):
                    pickled_value = pipe.lpop(self.key)
                    pipe.rpush(self.key, pickled_value)

        forward = n >= 0
        abs_n = abs(n)
        self._transaction(rotate_trans)

    # Operator methods

    def __add__(self, other):
        if not isinstance(other, (self.__class__, self._python_cls)):
            raise TypeError

        if self._same_redis(other, RedisCollection):
            return self._add_helper(
                other, use_redis=True, maxlen=self.maxlen
            )

        return self._add_helper(other, maxlen=self.maxlen)

    def __radd__(self, other):
        if not isinstance(other, (self.__class__, self._python_cls)):
            raise TypeError

        return self._add_helper(
            other, swap_args=True, maxlen=other.maxlen
        )

    def __iadd__(self, other):
        if not isinstance(other, (self.__class__, self._python_cls)):
            raise TypeError

        self.extend(other)
        return self

    def __mul__(self, times):
        if not isinstance(times, six.integer_types):
            raise TypeError

        return self._python_cls(self.__iter__(), self.maxlen) * times



# -*- coding: utf-8 -*-
from __future__ import division, print_function, unicode_literals

__title__ = 'redis-collections'
__version__ = '0.4.2'
__author__ = 'Honza Javorek'
__license__ = 'ISC'
__copyright__ = 'Copyright 2013-? Honza Javorek'


from .base import RedisCollection  # NOQA
from .dicts import DefaultDict, Dict, Counter  # NOQA
from .lists import Deque, List  # NOQA
from .sets import Set  # NOQA
from .sortedsets import SortedSetCounter  # NOQA
from .syncable import (  # NOQA
    LRUDict,
    SyncableDict,
    SyncableCounter,
    SyncableDeque,
    SyncableDefaultDict,
    SyncableList,
    SyncableSet,
)

__all__ = [
    'Counter',
    'DefaultDict',
    'Deque',
    'Dict',
    'List',
    'LRUDict',
    'RedisCollection',
    'Set',
    'SortedSetCounter',
    'SyncableDict',
    'SyncableCounter',
    'SyncableDefaultDict',
    'SyncableDeque',
    'SyncableList',
    'SyncableSet',
]



# -*- coding: utf-8 -*-
"""
sets
~~~~~

The `sets` module contains a standard collection, :class:`Set`, which is based
on Python's built-in set type.
Its elements are stored in a Redis `set <http://redis.io/commands#set>`_
structure.

"""
from __future__ import division, print_function, unicode_literals

import collections
from functools import reduce
import operator
import random

import six

from .base import RedisCollection


class Set(RedisCollection, collections.MutableSet):
    """
    Collection based on the built-in Python :class:`set` type.
    Items are stored in a Redis hash structure.
    See Python's `set documentation
    <https://docs.python.org/3/library/stdtypes.html#set>`_ for usage notes.
    """

    if six.PY2:
        _pickle = RedisCollection._pickle_2
        _unpickle = RedisCollection._unpickle_2
    else:
        _pickle = RedisCollection._pickle_3

    def __init__(self, *args, **kwargs):
        """
        Create a new Set object.

        If the first argument (*data*) is an iterable object, create the new
        Set with its elements as the initial data.

        :param data: Initial data.
        :type data: iterable
        :param redis: Redis client instance. If not provided, default Redis
                      connection is used.
        :type redis: :class:`redis.StrictRedis`
        :param key: Redis key for the collection. Collections with the same key
                    point to the same data. If not provided, a random
                    string is generated.
        :type key: str
        """
        data = args[0] if args else kwargs.pop('data', None)
        super(Set, self).__init__(**kwargs)

        if data:
            self.update(data)

    def _data(self, pipe=None):
        pipe = self.redis if pipe is None else pipe
        return (self._unpickle(x) for x in pipe.smembers(self.key))

    def _repr_data(self):
        items = (repr(v) for v in self.__iter__())
        return '{{{}}}'.format(', '.join(items))

    # Magic methods

    def __contains__(self, value, pipe=None):
        """Test for membership of *value* in the set."""
        pipe = self.redis if pipe is None else pipe
        return bool(pipe.sismember(self.key, self._pickle(value)))

    def __iter__(self, pipe=None):
        """Return an iterator over elements of the set."""
        pipe = self.redis if pipe is None else pipe
        return self._data(pipe)

    def __len__(self, pipe=None):
        """Return cardinality of the set."""
        pipe = self.redis if pipe is None else pipe
        return pipe.scard(self.key)

    # Named methods

    def add(self, value):
        """Add element *value* to the set."""
        # Raise TypeError if value is not hashable
        hash(value)

        self.redis.sadd(self.key, self._pickle(value))

    def copy(self, key=None):
        other = self.__class__(redis=self.redis, key=key)
        other.update(self)

        return other

    def clear(self, pipe=None):
        """Remove all elements from the set."""
        self._clear(pipe)

    def discard(self, value):
        """Remove element *value* from the set if it is present."""
        # Raise TypeError if value is not hashable
        hash(value)

        self.redis.srem(self.key, self._pickle(value))

    def isdisjoint(self, other):
        """
        Return ``True`` if the set has no elements in common with *other*.
        Sets are disjoint if and only if their intersection is the empty set.

        :param other: Any kind of iterable.
        :rtype: boolean
        """
        def isdisjoint_trans_pure(pipe):
            return not pipe.sinter(self.key, other.key)

        def isdisjoint_trans_mixed(pipe):
            self_values = set(self.__iter__(pipe))
            if use_redis:
                other_values = set(other.__iter__(pipe))
            else:
                other_values = set(other)

            return self_values.isdisjoint(other_values)

        if self._same_redis(other):
            return self._transaction(isdisjoint_trans_pure, other.key)
        if self._same_redis(other, RedisCollection):
            use_redis = True
            return self._transaction(isdisjoint_trans_mixed, other.key)

        use_redis = False
        return self._transaction(isdisjoint_trans_mixed)

    def pop(self):
        """
        Remove and return an arbitrary element from the set.
        Raises :exc:`KeyError` if the set is empty.
        """
        result = self.redis.spop(self.key)
        if result is None:
            raise KeyError

        return self._unpickle(result)

    def random_sample(self, k=1):
        """
        Return a *k* length list of unique elements chosen from the Set.
        Elements are not removed. Similar to :func:`random.sample` function
        from standard library.

        :param k: Size of the sample, defaults to 1.
        :rtype: :class:`list`

        .. note::
            This method is not available on the Python :class:`set`.
            When Redis version < 2.6 is being used the whole set is stored
            in memory and the sample is computed in Python.
        """
        # k == 0: no work to do
        if k == 0:
            results = []
        # k == 1: same behavior on all versions of Redis
        elif k == 1:
            results = [self.redis.srandmember(self.key)]
        # k != 1, Redis version >= 2.6: compute in Redis
        elif self.redis_version >= (2, 6, 0):
            results = self.redis.srandmember(self.key, k)
        # positive k, Redis version < 2.6: sample without replacement
        elif k > 1:
            seq = list(self.__iter__())
            return random.sample(seq, min(k, len(seq)))
        # negative k, Redis version < 2.6: sample with replacement
        else:
            seq = list(self.__iter__())
            return [random.choice(seq) for __ in six.moves.xrange(abs(k))]

        return [self._unpickle(x) for x in results]

    def remove(self, value):
        """
        Remove element *value* from the set. Raises :exc:`KeyError` if it
        is not contained in the set.
        """
        # Raise TypeError if value is not hashable
        hash(value)

        result = self.redis.srem(self.key, self._pickle(value))
        if not result:
            raise KeyError(value)

    def scan_elements(self):
        """
        Yield each of the elements from the collection, without pulling them
        all into memory.

        .. warning::
            This method is not available on the set collections provided
            by Python.

            This method may return the element multiple times.
            See the `Redis SCAN documentation
            <http://redis.io/commands/scan#scan-guarantees>`_ for details.
        """
        for x in self.redis.sscan_iter(self.key):
            yield self._unpickle(x)

    # Comparison and set operation helpers

    def _ge_helper(self, other, op, check_type=False):
        if check_type and not isinstance(other, collections.Set):
            raise TypeError

        def ge_trans_pure(pipe):
            if not op(self.__len__(pipe), other.__len__(pipe)):
                return False

            return not pipe.sdiff(other.key, self.key)

        def ge_trans_mixed(pipe):
            len_other = other.__len__(pipe) if use_redis else len(other)
            if not op(self.__len__(pipe), len_other):
                return False

            values = set(other.__iter__(pipe)) if use_redis else set(other)
            return all(self.__contains__(v, pipe=pipe) for v in values)

        if self._same_redis(other):
            return self._transaction(ge_trans_pure, other.key)
        if self._same_redis(other, RedisCollection):
            use_redis = True
            return self._transaction(ge_trans_mixed, other.key)

        use_redis = False
        return self._transaction(ge_trans_mixed)

    def _le_helper(self, other, op, check_type=False):
        if check_type and not isinstance(other, collections.Set):
            raise TypeError

        def le_trans_pure(pipe):
            if not op(self.__len__(pipe), other.__len__(pipe)):
                return False

            return not pipe.sdiff(self.key, other.key)

        def le_trans_mixed(pipe):
            len_other = other.__len__(pipe) if use_redis else len(other)
            if not op(self.__len__(pipe), len_other):
                return False

            values = set(other.__iter__(pipe)) if use_redis else set(other)
            return all(v in values for v in self.__iter__(pipe))

        if self._same_redis(other):
            return self._transaction(le_trans_pure, other.key)
        if self._same_redis(other, RedisCollection):
            use_redis = True
            return self._transaction(le_trans_mixed, other.key)

        use_redis = False
        return self._transaction(le_trans_mixed)

    def _op_update_helper(
        self, others, op, redis_op, update=False, check_type=False
    ):
        if (
            check_type and
            not all(isinstance(x, collections.Set) for x in others)
        ):
                raise TypeError

        def op_update_trans_pure(pipe):
            method = getattr(pipe, redis_op)
            if not update:
                result = method(self.key, *other_keys)
                return {self._unpickle(x) for x in result}

            temp_key = self._create_key()
            pipe.multi()
            method(temp_key, self.key, *other_keys)
            pipe.rename(temp_key, self.key)

        def op_update_trans_mixed(pipe):
            self_values = set(self.__iter__(pipe))
            other_values = []
            for other in others:
                if isinstance(other, RedisCollection):
                    other_values.append(set(other.__iter__(pipe)))
                else:
                    other_values.append(set(other))

            if not update:
                return reduce(op, other_values, self_values)

            new_values = reduce(op, other_values, self_values)
            pipe.multi()
            pipe.delete(self.key)
            for v in new_values:
                pipe.sadd(self.key, self._pickle(v))

        other_keys = []
        all_redis_sets = True
        for other in others:
            if self._same_redis(other):
                other_keys.append(other.key)
            elif self._same_redis(other, RedisCollection):
                other_keys.append(other.key)
                all_redis_sets = False
            else:
                all_redis_sets = False

        if all_redis_sets:
            return self._transaction(op_update_trans_pure, *other_keys)

        return self._transaction(op_update_trans_mixed, *other_keys)

    def _rop_helper(self, other, op):
        if not isinstance(other, collections.Set):
            raise TypeError

        return op(set(other), set(self.__iter__()))

    def _xor_helper(self, other, update=False, check_type=False):
        if check_type and not isinstance(other, collections.Set):
            raise TypeError

        def xor_trans_pure(pipe):
            diff_1_key = self._create_key()
            pipe.sdiffstore(diff_1_key, self.key, other.key)

            diff_2_key = self._create_key()
            pipe.sdiffstore(diff_2_key, other.key, self.key)

            if update:
                pipe.sunionstore(self.key, diff_1_key, diff_2_key)
                ret = None
            else:
                ret = pipe.sunion(diff_1_key, diff_2_key)
                ret = {self._unpickle(x) for x in ret}

            pipe.delete(diff_1_key, diff_2_key)

            return ret

        def xor_trans_mixed(pipe):
            self_values = set(self.__iter__(pipe))
            if use_redis:
                other_values = set(other.__iter__(pipe))
            else:
                other_values = set(other)

            result = self_values ^ other_values

            if update:
                pipe.delete(self.key)
                pipe.sadd(self.key, *(self._pickle(x) for x in result))
                return None

            return result

        if self._same_redis(other):
            return self._transaction(xor_trans_pure, other.key)
        elif self._same_redis(other, RedisCollection):
            use_redis = True
            return self._transaction(xor_trans_mixed, other.key)

        use_redis = False
        return self._transaction(xor_trans_mixed)

    # Intersection

    def __and__(self, other):
        return self._op_update_helper(
            (other,), operator.and_, 'sinter', check_type=True
        )

    def __rand__(self, other):
        return self._rop_helper(other, operator.and_)

    def __iand__(self, other):
        self._op_update_helper(
            (other,),
            operator.and_,
            'sinterstore',
            update=True,
            check_type=True,
        )
        return self

    def intersection(self, *others):
        """
        Return a new set with elements common to the set and all *others*.

        :param others: Iterables, each one as a single positional argument.
        :rtype: :class:`set`

        .. note::
            The same behavior as at :func:`union` applies.
        """
        return self._op_update_helper(tuple(others), operator.and_, 'sinter')

    def intersection_update(self, *others):
        """
        Update the set, keeping only elements found in it and all *others*.

        :param others: Iterables, each one as a single positional argument.
        :rtype: None

        .. note::
            The same behavior as at :func:`difference_update` applies.
        """
        return self._op_update_helper(
            tuple(others), operator.and_, 'sinterstore', update=True
        )

    # Comparison

    def __ge__(self, other):
        return self._ge_helper(other, operator.ge, check_type=True)

    def issuperset(self, other):
        """
        Test whether every element in other is in the set.

        :param other: Any kind of iterable.
        :rtype: boolean
        """
        return self._ge_helper(other, operator.ge)

    def __gt__(self, other):
        return self._ge_helper(other, operator.gt, check_type=True)

    def __eq__(self, other):
        return self._le_helper(other, operator.eq, check_type=True)

    def __le__(self, other):
        return self._le_helper(other, operator.le, check_type=True)

    def issubset(self, other):
        """
        Test whether every element in the set is in *other*.

        :param other: Any kind of iterable.
        :rtype: boolean
        """
        return self._le_helper(other, operator.le)

    def __lt__(self, other):
        return self._le_helper(other, operator.lt)

    # Union

    def __or__(self, other):
        return self._op_update_helper(
            (other,), operator.or_, 'sunion', check_type=True
        )

    def __ror__(self, other):
        return self._rop_helper(other, operator.or_)

    def __ior__(self, other):
        self._op_update_helper(
            (other,), operator.or_, 'sunionstore', update=True, check_type=True
        )
        return self

    def union(self, *others):
        """
        Return a new set with elements from the set and all *others*.

        :param others: Iterables, each one as a single positional argument.
        :rtype: :class:`set`

        .. note::
            If all *others* are :class:`Set` instances, the operation
            is performed completely in Redis. Otherwise, values are retrieved
            from Redis and the operation is performed in Python.
        """
        return self._op_update_helper(tuple(others), operator.or_, 'sunion')

    def update(self, *others):
        """
        Update the set, adding elements from all *others*.

        :param others: Iterables, each one as a single positional argument.
        :rtype: None

        .. note::
            If all *others* are :class:`Set` instances, the operation
            is performed completely in Redis. Otherwise, values are retrieved
            from Redis and the operation is performed in Python.
        """
        return self._op_update_helper(
            tuple(others), operator.or_, 'sunionstore', update=True
        )

    # Difference

    def __sub__(self, other):
        return self._op_update_helper(
            (other,), operator.sub, 'sdiff', check_type=True
        )

    def __rsub__(self, other):
        return self._rop_helper(other, operator.sub)

    def __isub__(self, other):
        self._op_update_helper(
            (other,), operator.sub, 'sdiffstore', update=True, check_type=True
        )
        return self

    def difference(self, *others):
        """
        Return a new set with elements in the set that are not in the *others*.

        :param others: Iterables, each one as a single positional argument.
        :rtype: :class:`set`

        .. note::
            The same behavior as at :func:`union` applies.
        """
        return self._op_update_helper(tuple(others), operator.sub, 'sdiff')

    def difference_update(self, *others):
        """
        Update the set, removing elements found in *others*.

        :param others: Iterables, each one as a single positional argument.
        :rtype: None

        .. note::
            The same behavior as at :func:`update` applies.
        """
        return self._op_update_helper(
            tuple(others), operator.sub, 'sdiffstore', update=True
        )

    # Symmetric difference

    def __xor__(self, other):
        return self._xor_helper(other, check_type=True)

    def __ixor__(self, other):
        self._xor_helper(other, update=True, check_type=True)
        return self

    def symmetric_difference(self, other):
        """
        Return a new set with elements in either the set or *other* but not
        both.

        :param other: Any kind of iterable.
        :rtype: :class:`set`

        .. note::
            The same behavior as at :func:`union` applies.
        """
        return self._xor_helper(other)

    def symmetric_difference_update(self, other):
        """
        Update the set, keeping only elements found in either set, but not
        in both.

        :param other: Any kind of iterable.
        :rtype: None

        .. note::
            The same behavior as at :func:`update` applies.
        """
        self._xor_helper(other, update=True)
        return self



# -*- coding: utf-8 -*-
"""
dicts
~~~~~

The `dicts` module contains standard collections based on Python dictionaries.
Included collections are :class:`Dict`, :class:`Counter`, and
:class:`DefaultDict`.
Each collection stores its items in a Redis
`hash <http://redis.io/commands#hash>`_ structure.

.. note::
    If you need to store mutable values like :class:`list`\s or :class:`set`\s
    in a collection, be sure to enable ``writeback``.
    See :ref:`Synchronization` for more information.

    When storing numeric types (e.g. :class:`float`) as keys, be aware that
    these collections behave slightly differently from standard Python
    dictionaries.
    See :ref:`Hashing` for more information.

"""
from __future__ import division, print_function, unicode_literals

import collections
import operator

import six

from .base import RedisCollection


class Dict(RedisCollection, collections.MutableMapping):
    """
    Collection based on the built-in Python :class:`dict` type.
    Items are stored in a Redis hash structure.
    See Python's `dict documentation
    <https://docs.python.org/3/library/stdtypes.html#mapping-types-dict>`_ for
    usage notes.

    The :func:`viewitems`, :func:`viewkeys`, and :func:`viewvalues` methods
    from Python 2.7's dictionary type are not implemented.
    """

    if six.PY2:
        _pickle_key = RedisCollection._pickle_2
        _unpickle_key = RedisCollection._unpickle_2
    else:
        _pickle_key = RedisCollection._pickle_3
        _unpickle_key = RedisCollection._unpickle

    _pickle_value = RedisCollection._pickle_3

    class __missing_value(object):
        def __repr__(self):
            # Specified here so that the documentation shows a useful string
            # for methods that take __marker as a keyword argument
            return '<missing value>'
    __marker = __missing_value()

    def __init__(self, *args, **kwargs):
        """
        Create a new Dict object.

        If the first argument (*data*) is another mapping type, create the new
        Dict with its items as the initial data.
        Or, If the first argument is an iterable of ``(key, value)`` pairs,
        create the new Dict with those items as the initial data.

        Unlike Python's built-in :class:`dict` type, initial items cannot be
        set using keyword arguments.
        Keyword arguments are passed to the :class:`RedisCollection`
        constructor.

        :param data: Initial data.
        :type data: iterable or mapping

        :param redis: Redis client instance. If not provided, default Redis
                      connection is used.
        :type redis: :class:`redis.StrictRedis`

        :param key: Redis key for the collection. Collections with the same key
                    point to the same data. If not provided, a random
                    string is generated.
        :type key: str

        :param writeback: If ``True``, keep a local cache of changes for
                          storing modifications to mutable values. Changes will
                          be written to Redis after calling the ``sync``
                          method.
        :type writeback: bool
        """
        data = args[0] if args else kwargs.pop('data', None)
        writeback = kwargs.pop('writeback', False)
        super(Dict, self).__init__(**kwargs)

        self.writeback = writeback
        self.cache = {}

        if data:
            self.update(data)

    def __len__(self, pipe=None):
        """Return the number of items in the dictionary."""
        pipe = self.redis if pipe is None else pipe
        return pipe.hlen(self.key)

    def __iter__(self, pipe=None):
        """Return an iterator over the keys of the dictionary."""
        pipe = self.redis if pipe is None else pipe
        for k, v in six.iteritems(self._data(pipe)):
            yield k

    def __contains__(self, key):
        """Return ``True`` if *key* is present, else ``False``."""
        pickled_key = self._pickle_key(key)
        return bool(self.redis.hexists(self.key, pickled_key))

    def __eq__(self, other):
        if not isinstance(other, collections.Mapping):
            return False

        def eq_trans(pipe):
            self_items = self.iteritems(pipe)
            other_items = other.items(pipe) if use_redis else other.items()

            return dict(self_items) == dict(other_items)

        if self._same_redis(other, RedisCollection):
            use_redis = True
            return self._transaction(eq_trans, other.key)
        else:
            use_redis = False
            return self._transaction(eq_trans)

    def getmany(self, *keys):
        """
        Return a list of values corresponding to the keys in the iterable of
        *keys*.
        If a key is not present in the collection, its corresponding value will
        be :obj:`None`.

        .. note::
            This method is not implemented by standard Python dictionary
            classes.
        """
        pickled_keys = (self._pickle_key(k) for k in keys)
        pickled_values = self.redis.hmget(self.key, *pickled_keys)

        ret = []
        for k, v in six.moves.zip(keys, pickled_values):
            value = self.cache.get(k, self._unpickle(v))
            ret.append(value)

        return ret

    def __getitem__(self, key):
        """Return the item of dictionary with key *key*. Raises a
        :exc:`KeyError` if key is not in the map.

        If a subclass of :class:`Dict` defines a method :func:`__missing__`,
        and *key* is not present, the ``d[key]`` operation calls that
        method with the key *key* as argument.
        The ``d[key]`` operation then returns or raises whatever is returned
        or raised by the ``__missing__(key)`` call.
        """
        try:
            value = self.cache[key]
        except KeyError:
            pickled_key = self._pickle_key(key)
            pickled_value = self.redis.hget(self.key, pickled_key)
            if pickled_value is None:
                if hasattr(self, '__missing__'):
                    return self.__missing__(key)
                raise KeyError(key)

            value = self._unpickle(pickled_value)

        if self.writeback:
            self.cache[key] = value
        return value

    def __setitem__(self, key, value):
        """Set ``d[key]`` to *value*."""
        pickled_key = self._pickle_key(key)
        pickled_value = self._pickle_value(value)
        self.redis.hset(self.key, pickled_key, pickled_value)

        if self.writeback:
            self.cache[key] = value

    def __delitem__(self, key):
        """Remove ``d[key]`` from dictionary.
        Raises a :func:`KeyError` if *key* is not in the map.
        """
        pickled_key = self._pickle_key(key)
        deleted_count = self.redis.hdel(self.key, pickled_key)
        if not deleted_count:
            raise KeyError(key)

        self.cache.pop(key, None)

    def _data(self, pipe=None):
        """
        Returns a Python dictionary with the same values as this object
        (without checking the local cache).
        """
        pipe = self.redis if pipe is None else pipe
        items = six.iteritems(pipe.hgetall(self.key))

        return {self._unpickle_key(k): self._unpickle(v) for k, v in items}

    def items(self, pipe=None):
        """Return a copy of the dictionary's list of ``(key, value)`` pairs."""
        return list(self.iteritems(pipe))

    def iteritems(self, pipe=None):
        """Return an iterator over the dictionary's ``(key, value)`` pairs."""
        pipe = self.redis if pipe is None else pipe
        for k, v in six.iteritems(self._data(pipe)):
            yield k, self.cache.get(k, v)

    def keys(self):
        """Return a copy of the dictionary's list of keys."""
        return list(self.__iter__())

    def iter(self):
        """Return an iterator over the keys of the dictionary.
        This is a shortcut for :func:`iterkeys()`.
        """
        return self.__iter__()

    def iterkeys(self):
        """Return an iterator over the dictionary's keys."""
        return self.__iter__()

    def values(self):
        """Return a copy of the dictionary's list of values."""
        return [v for k, v in self.iteritems()]

    def itervalues(self):
        """Return an iterator over the dictionary's values."""
        return (v for k, v in self.iteritems())

    def pop(self, key, default=__marker):
        """If *key* is in the dictionary, remove it and return its value,
        else return *default*. If *default* is not given and *key* is not
        in the dictionary, a :exc:`KeyError` is raised.
        """
        pickled_key = self._pickle_key(key)

        if key in self.cache:
            self.redis.hdel(self.key, pickled_key)
            return self.cache.pop(key)

        def pop_trans(pipe):
            pickled_value = pipe.hget(self.key, pickled_key)
            if pickled_value is None:
                if default is self.__marker:
                    raise KeyError(key)
                return default

            pipe.hdel(self.key, pickled_key)
            return self._unpickle(pickled_value)

        value = self._transaction(pop_trans)
        self.cache.pop(key, None)

        return value

    def popitem(self):
        """Remove and return an arbitrary ``(key, value)`` pair from
        the dictionary.

        :func:`popitem` is useful to destructively iterate over
        a dictionary, as often used in set algorithms. If
        the dictionary is empty, calling :func:`popitem` raises
        a :exc:`KeyError`.
        """
        def popitem_trans(pipe):
            try:
                pickled_key = pipe.hkeys(self.key)[0]
            except IndexError:
                raise KeyError

            # pop its value
            pipe.multi()
            pipe.hget(self.key, pickled_key)
            pipe.hdel(self.key, pickled_key)
            pickled_value, __ = pipe.execute()

            return (
                self._unpickle_key(pickled_key), self._unpickle(pickled_value)
            )

        key, value = self._transaction(popitem_trans)

        return key, self.cache.pop(key, value)

    def setdefault(self, key, default=None):
        """If *key* is in the dictionary, return its value.
        If not, insert *key* with a value of *default* and
        return *default*. *default* defaults to :obj:`None`.
        """
        if key in self.cache:
            return self.cache[key]

        def setdefault_trans(pipe):
            pickled_key = self._pickle_key(key)

            pipe.multi()
            pipe.hsetnx(self.key, pickled_key, self._pickle_value(default))
            pipe.hget(self.key, pickled_key)

            __, pickled_value = pipe.execute()

            return self._unpickle(pickled_value)

        value = self._transaction(setdefault_trans)

        if self.writeback:
            self.cache[key] = value
        return value

    def _update_helper(self, other, use_redis=False):
        def _update_helper_trans(pipe):
            data = {}

            if isinstance(other, Dict):
                data.update(other.iteritems(pipe))
            elif isinstance(other, RedisCollection):
                data.update(other.__iter__(pipe))
            else:
                data.update(other)

            pickled_data = {}
            for k, v in six.iteritems(data):
                pickled_data[self._pickle_key(k)] = self._pickle_value(v)

            if pickled_data:
                pipe.hmset(self.key, pickled_data)

            if self.writeback:
                self.cache.update(data)

        if use_redis:
            self._transaction(_update_helper_trans, other.key)
        else:
            self._transaction(_update_helper_trans)

    def update(self, other=None, **kwargs):
        """Update the dictionary with the key/value pairs from *other*,
        overwriting existing keys. Return :obj:`None`.

        :func:`update` accepts either another dictionary object or
        an iterable of key/value pairs (as tuples or other iterables
        of length two). If keyword arguments are specified, the
        dictionary is then updated with those key/value pairs:
        ``d.update(red=1, blue=2)``.
        """
        if other is not None:
            if self._same_redis(other, RedisCollection):
                self._update_helper(other, use_redis=True)
            elif hasattr(other, 'keys'):
                self._update_helper(other)
            else:
                self._update_helper({k: v for k, v in other})

        if kwargs:
            self._update_helper(kwargs)

    def copy(self, key=None):
        """
        Return a new collection with the same items as this one.
        If *key* is specified, create the new collection with the given
        Redis key.
        """
        other = self.__class__(redis=self.redis, key=key)
        other.update(self)

        return other

    def clear(self, pipe=None):
        self._clear(pipe)

        if self.writeback:
            self.cache.clear()

    @classmethod
    def fromkeys(cls, seq, value=None, **kwargs):
        """Create a new dictionary with keys from *seq* and values set to
        *value*.

        .. note::
            :func:`fromkeys` is a class method that returns a new dictionary.
            It is possible to specify additional keyword arguments to be passed
            to :func:`__init__` of the new object.
        """
        values = ((key, value) for key in seq)
        return cls(values, **kwargs)

    def scan_items(self):
        """
        Yield each of the ``(key, value)`` pairs from the collection, without
        pulling them all into memory.

        .. warning::
            This method is not available on the dictionary collections provided
            by Python.

            This method may return the same (key, value) pair multiple times.
            See the `Redis SCAN documentation
            <http://redis.io/commands/scan#scan-guarantees>`_ for details.
        """
        for k, v in self.redis.hscan_iter(self.key):
            yield self._unpickle_key(k), self._unpickle(v)

    def _repr_data(self):
        items = ('{}: {}'.format(repr(k), repr(v)) for k, v in self.items())
        return '{{{}}}'.format(', '.join(items))

    def sync(self):
        self.writeback = False
        self._update_helper(self.cache)
        self.cache = {}
        self.writeback = True


class Counter(Dict):
    """
    Collection based on the Python standard library's
    :class:`collections.Counter` type.
    Items are stored in a Redis hash structure.
    See Python's `Counter documentation
    <http://docs.python.org/2/library/collections.html#collections.Counter>`_
    for usage notes.

    Counter inherits from Dict, so see its API documentation for information
    on other methods.

    The :func:`viewitems`, :func:`viewkeys`, and :func:`viewvalues` methods
    from Python 2.7's Counter type are not implemented.
    """

    def __init__(self, *args, **kwargs):
        """
        Create a new Counter object.

        If the first argument (*data*) is another mapping type, create the new
        Counter with the counts of the input items as the initial data.
        Or, If the first argument is an iterable of ``(key, value)`` pairs,
        create the new Counter with those items as the initial data.

        Unlike Python's standard :class:`collections.Counter` type,
        initial items cannot be set using keyword arguments.
        Keyword arguments are passed to the :class:`RedisCollection`
        constructor.

        :param data: Initial data.
        :type data: iterable or mapping

        :param redis: Redis client instance. If not provided, default Redis
                      connection is used.
        :type redis: :class:`redis.StrictRedis`

        :param key: Redis key for the collection. Collections with the same key
                    point to the same data. If not provided, a random
                    string is generated.
        :type key: str

        :param writeback: If ``True``, keep a local cache of changes for
                          storing modifications to mutable values. Changes will
                          be written to Redis after calling the ``sync``
                          method.
        :type writeback: bool
        """
        super(Counter, self).__init__(*args, **kwargs)

    def __missing__(self, key):
        return 0

    def most_common(self, n=None):
        """Return a list of the *n* most common elements and their counts
        from the most common to the least. If *n* is not specified,
        :func:`most_common` returns *all* elements in the counter.
        Elements with equal counts are ordered arbitrarily.
        """
        return collections.Counter(self).most_common(n)

    def elements(self, n=None):
        """Return an iterator over elements repeating each as many times as
        its count. Elements are returned in arbitrary order. If an element's
        count is less than one, :func:`elements` will ignore it.
        """
        return collections.Counter(self).elements()

    @classmethod
    def fromkeys(cls, iterable, v=None):
        raise NotImplementedError(
            'Counter.fromkeys() is undefined.  Use Counter(iterable) instead.'
        )

    def _update_helper(self, other, op, use_redis=False):
        def _update_helper_trans(pipe):
            data = {}

            if isinstance(other, Dict):
                data.update(other.iteritems(pipe))
            elif isinstance(other, RedisCollection):
                data.update(collections.Counter(other.__iter__(pipe)))
            else:
                data.update(other)

            pickled_data = {}
            for k, v in six.iteritems(data):
                pickled_key = self._pickle_key(k)
                pickled_value = self._pickle_value(op(self.get(k, 0), v))
                pickled_data[pickled_key] = pickled_value

            if pickled_data:
                pipe.hmset(self.key, pickled_data)

            if self.writeback:
                self.cache.update(data)

        if use_redis:
            self._transaction(_update_helper_trans, other.key)
        else:
            self._transaction(_update_helper_trans)

    def update(self, other=None, **kwargs):
        """Elements are counted from an *iterable* or added-in from another
        *mapping* (or counter). Like :func:`dict.update` but adds counts
        instead of replacing them. Also, the *iterable* is expected to be
        a sequence of elements, not a sequence of ``(key, value)`` pairs.
        """
        if other is not None:
            if self._same_redis(other, RedisCollection):
                self._update_helper(other, operator.add, use_redis=True)
            elif hasattr(other, 'keys'):
                self._update_helper(other, operator.add)
            else:
                self._update_helper(collections.Counter(other), operator.add)

        if kwargs:
            self._update_helper(kwargs, operator.add)

    def subtract(self, other=None, **kwargs):
        """Elements are subtracted from an *iterable* or from another
        *mapping* (or counter). Like :func:`dict.update` but subtracts
        counts instead of replacing them.
        """
        if other is not None:
            if self._same_redis(other, RedisCollection):
                self._update_helper(other, operator.sub, use_redis=True)
            elif hasattr(other, 'keys'):
                self._update_helper(other, operator.sub)
            else:
                self._update_helper(collections.Counter(other), operator.sub)

        if kwargs:
            self._update_helper(kwargs, operator.sub)

    def __delitem__(self, key):
        """Like :func:`dict.__delitem__`, but does not raise KeyError for
        missing values.
        """
        try:
            super(Counter, self).__delitem__(key)
        except KeyError:
            pass

    def _op_helper(self, other, op, swap_args=False, inplace=False):
        def op_trans(pipe):
            # Get a collections.Counter copy of `self`
            self_counter = collections.Counter(
                {k: v for k, v in self.iteritems(pipe=pipe)}
            )

            # If `other` is also Redis-backed we'll want to pull its values
            # with the same transaction-provided pipeline as for `self`.
            if isinstance(other, Dict):
                other_counter = collections.Counter(
                    {k: v for k, v in other.iteritems(pipe=pipe)}
                )
            else:
                other_counter = other

            # Unary case
            if other is None:
                result = op(self_counter)
            # Reversed case
            elif swap_args:
                result = op(other_counter, self_counter)
            # Normal case
            else:
                result = op(self_counter, other_counter)

            # If we're not updating `self`, we're finished
            if not inplace:
                return result

            # Otherwise we need to update `self` in this transaction
            pickled_data = {}
            for key, value in six.iteritems(result):
                pickled_key = self._pickle_key(key)
                pickled_value = self._pickle_value(value)
                pickled_data[pickled_key] = pickled_value

            pipe.multi()
            pipe.delete(self.key)
            if pickled_data:
                pipe.hmset(self.key, pickled_data)

        if other is None:
            result = self._transaction(op_trans, None)
        elif self._same_redis(other, RedisCollection):
            result = self._transaction(op_trans, other.key)
        elif isinstance(other, collections.Counter):
            result = self._transaction(op_trans)
        else:
            raise TypeError('Unsupported type {}'.format(type(other)))

        return self if inplace else result

    def __add__(self, other):
        return self._op_helper(other, operator.add)

    def __radd__(self, other):
        return self._op_helper(other, operator.add, swap_args=True)

    def __sub__(self, other):
        return self._op_helper(other, operator.sub)

    def __rsub__(self, other):
        return self._op_helper(other, operator.sub, swap_args=True)

    def __or__(self, other):
        return self._op_helper(other, operator.or_)

    def __ror__(self, other):
        return self._op_helper(other, operator.or_, swap_args=True)

    def __and__(self, other):
        return self._op_helper(other, operator.and_)

    def __rand__(self, other):
        return self._op_helper(other, operator.and_, swap_args=True)

    def __iadd__(self, other):
        return self._op_helper(other, operator.add, inplace=True)

    def __isub__(self, other):
        return self._op_helper(other, operator.sub, inplace=True)

    def __ior__(self, other):
        return self._op_helper(other, operator.ior, inplace=True)

    def __iand__(self, other):
        return self._op_helper(other, operator.iand, inplace=True)

    def __pos__(self):
        return self._op_helper(None, operator.pos)

    def __neg__(self):
        return self._op_helper(None, operator.neg)


class DefaultDict(Dict):
    """
    Collection based on the Python standard library's
    :class:`collections.defaultdict` type.
    Items are stored in a Redis hash structure.
    See Python's `defaultdict documentation
    <https://docs.python.org/3/library/collections.html#collections.defaultdict>`_
    for usage notes.

    DefaultDict inherits from Dict, so see its API documentation for
    information on other methods.

    The :func:`viewitems`, :func:`viewkeys`, and :func:`viewvalues` methods
    from Python 2.7's Counter type are not implemented.
    """

    def __init__(self, *args, **kwargs):
        """
        Create a new DefaultDict object.

        The first argument provides the initial value for the
        ``default_factory`` attribute; it defaults to ``None``.
        All other arguments are passed to the ``Dict`` constructor.

        Unlike Python's standard :class:`collections.defaultdict` type,
        initial items cannot be set using keyword arguments.
        Keyword arguments are passed to the :class:`RedisCollection`
        constructor via the ``Dict`` constructor.

        :param default_factory: Used to provide default values for missing
                                keys.
        :type default_factory: callable or None
        :param data: Initial data.
        :type data: iterable or mapping
        :param redis: Redis client instance. If not provided, default Redis
                      connection is used.
        :type redis: :class:`redis.StrictRedis`
        :param key: Redis key for the collection. Collections with the same key
                    point to the same data. If not provided, a random
                    string is generated.
        :type key: str
        """
        kwargs.setdefault('writeback', True)
        if args:
            default_factory = args[0]
            args = args[1:]
        else:
            default_factory = None

        super(DefaultDict, self).__init__(*args, **kwargs)

        if default_factory is None:
            pass
        elif not callable(default_factory):
            raise TypeError('first argument must be callable or None')
        self.default_factory = default_factory

    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)

        value = self.default_factory()
        self[key] = value
        return value

    def copy(self, key=None):
        """
        Return a new collection with the same items as this one.
        If *key* is specified, create the new collection with the given
        Redis key.
        """
        other = self.__class__(self.default_factory, redis=self.redis, key=key)
        other.update(self)

        return other



# -*- coding: utf-8 -*-
"""
base
~~~~
"""
from __future__ import division, print_function, unicode_literals

import abc
from decimal import Decimal
from fractions import Fraction
import uuid

# We use pickle instead of cPickle on Python 2 intentionally, see
# http://bugs.python.org/issue5518
import pickle

import redis
import six

NUMERIC_TYPES = six.integer_types + (float, Decimal, Fraction, complex)


@six.add_metaclass(abc.ABCMeta)
class RedisCollection(object):
    """Abstract class providing backend functionality for all the other
    Redis collections.
    """

    not_impl_msg = ('Cannot be implemented efficiently or atomically '
                    'due to limitations in Redis command set.')

    @abc.abstractmethod
    def __init__(self, redis=None, key=None):
        """
        :param data: Initial data.
        :param redis: Redis client instance. If not provided, a new Redis
                      connection is created.
        :type redis: :class:`redis.StrictRedis`
        :param key: The key at which the collection will be stored in Redis.
                    Collections with the same key point to the same data.
                    If not provided a random key is generated.
        :type key: str
        """
        #: Redis client instance. :class:`StrictRedis` object with default
        #: connection settings is used if not set by :func:`__init__`.
        self.redis = redis or self._create_redis()
        self._redis_version = None

        #: Redis key of the collection.
        self.key = key or self._create_key()

    def _create_redis(self):
        """
        Creates a new Redis connection when none is specified during
        initialization.

        :rtype: :class:`redis.StrictRedis`
        """
        return redis.StrictRedis()

    def _create_key(self):
        """
        Creates a random Redis key for storing this collection's data.

        :rtype: string

        .. note::
            :func:`uuid.uuid4` is used. If you are not satisfied with its
            `collision
            probability <http://stackoverflow.com/a/786541/325365>`_,
            make your own implementation by subclassing and overriding this
            method.
        """
        return uuid.uuid4().hex

    @abc.abstractmethod
    def _data(self, pipe=None):
        """Helper for getting the collection's data within a transaction.

        :param pipe: Redis pipe in case creation is performed as a part
                     of transaction.
        :type pipe: :class:`redis.client.StrictPipeline` or
                    :class:`redis.client.StrictRedis`
        """

    def _pickle(self, data):
        """Converts given data to a bytes string.

        :param data: Data to be serialized.
        :type data: anything serializable
        :rtype: bytes
        """
        return pickle.dumps(data)

    def _pickle_2(self, data):
        # On Python 2 some values of the str and unicode types have the same
        # hash, are equal to each other, but nonetheless pickle to different
        # byte strings. This method encodes unicode types to str to help match
        # Python's behavior.
        # The length of {b'a', u'a'} is 1 on Python 2.x and 2 on Python 3.x
        if isinstance(data, six.text_type):
            data = data.encode('utf-8')

        return self._pickle_3(data)

    def _pickle_3(self, data):
        # Several numeric types are equal, have the same hash, but nonetheless
        # pickle to different byte strings. This method reduces them down to
        # integers to help match with Python's behavior.
        # len({1.0, 1, complex(1, 0)}) == 1
        if isinstance(data, complex):
            int_data = int(data.real)
            if data == int_data:
                data = int_data
        elif isinstance(data, NUMERIC_TYPES):
            int_data = int(data)
            if data == int_data:
                data = int_data

        return pickle.dumps(data)

    def _unpickle(self, pickled_data):
        """Convert *pickled_data* to a Python object and return it.

        :param pickled_data: Serialized data.
        :type pickled_data: bytes
        :rtype: anything serializable
        """
        return pickle.loads(pickled_data) if pickled_data else None

    def _unpickle_2(self, string):
        # Because we encoded text data in the pickle method, we should decode
        # it on the way back out
        data = pickle.loads(string) if string else None
        if isinstance(data, six.binary_type):
            try:
                data = data.decode('utf-8')
            except UnicodeDecodeError:
                pass

        return data

    def _clear(self, pipe=None):
        """Helper for clear operations.

        :param pipe: Redis pipe in case update is performed as a part
                     of transaction.
        :type pipe: :class:`redis.client.StrictPipeline` or
                    :class:`redis.client.StrictRedis`
        """
        redis = self.redis if pipe is None else pipe
        redis.delete(self.key)

    @property
    def redis_version(self):
        # Set the Redis version if it's not already set.
        if self._redis_version is None:
            self._redis_version = tuple(
                int(x) for x in self.redis.info()['redis_version'].split('.')
            )

        return self._redis_version

    def _same_redis(self, other, cls=None):
        cls = cls or self.__class__
        if not isinstance(other, cls):
            return False

        self_kwargs = self.redis.connection_pool.connection_kwargs
        other_kwargs = other.redis.connection_pool.connection_kwargs

        return (
            self_kwargs.get('host') == other_kwargs.get('host') and
            self_kwargs.get('port') == other_kwargs.get('port') and
            self_kwargs.get('path') == other_kwargs.get('path') and
            self_kwargs.get('db', 0) == other_kwargs.get('db', 0)
        )

    def _normalize_index(self, index, pipe=None):
        """Convert negative indexes into their positive equivalents."""
        pipe = self.redis if pipe is None else pipe
        len_self = self.__len__(pipe)
        positive_index = index if index >= 0 else len_self + index

        return len_self, positive_index

    def _normalize_slice(self, index, pipe=None):
        """Given a :obj:`slice` *index*, return a 4-tuple
        ``(start, stop, step, fowrward)``. The first three items can be used
        with the ``range`` function to retrieve the values associated with the
        slice; the last item indicates the direction.
        """
        if index.step == 0:
            raise ValueError
        pipe = self.redis if pipe is None else pipe

        len_self = self.__len__(pipe)

        step = index.step or 1
        forward = step > 0
        step = abs(step)

        if index.start is None:
            start = 0 if forward else len_self - 1
        elif index.start < 0:
            start = max(len_self + index.start, 0)
        else:
            start = min(index.start, len_self)

        if index.stop is None:
            stop = len_self if forward else -1
        elif index.stop < 0:
            stop = max(len_self + index.stop, 0)
        else:
            stop = min(index.stop, len_self)

        if not forward:
            start, stop = min(stop + 1, len_self), min(start + 1, len_self)

        return start, stop, step, forward, len_self

    def _transaction(self, fn, *extra_keys):
        """Helper simplifying code within watched transaction.

        Takes *fn*, function treated as a transaction. Returns whatever
        *fn* returns. ``self.key`` is watched. *fn* takes *pipe* as the
        only argument.

        :param fn: Closure treated as a transaction.
        :type fn: function *fn(pipe)*
        :param extra_keys: Optional list of additional keys to watch.
        :type extra_keys: list
        :rtype: whatever *fn* returns
        """
        results = []

        def trans(pipe):
            results.append(fn(pipe))

        self.redis.transaction(trans, self.key, *extra_keys)
        return results[0]

    def __enter__(self):
        self.writeback = True
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.sync()

    def sync(self):
        pass

    @abc.abstractmethod
    def _repr_data(self):
        """
        Abstract method for subclasses to implement.
        Return a string appropriate for displaying the contents of the
        collection. Called by __repr__.
        """

    def __repr__(self):
        cls_name = self.__class__.__name__
        data = self._repr_data()
        return '<redis_collections.%s at %s %s>' % (cls_name, self.key, data)



# -*- coding: utf-8 -*-
"""
syncable
~~~~~~~~

Persistent Python collections that can be written to and read from Redis.
The collections are kept in memory, so their operations run as fast as their
standard counterparts'.

Use in a ``with`` block to automatically sync to Redis after the block
executes, or call the :func:`sync` method explicitly.

    >>> with SyncableDict() as D:
    ...     D['one'] = 1
    ...
    >>> D  # Contents are available locally and are stored in Redis
    {'one': 1}
    >>> D['two'] = 2  # Changes are available locally, but not in Redis...
    >>> D.sync()  # ...until sync is called.

If you specify a ``key`` pointing to an existing collection, its contents will
be loaded.

    >>> D.key
    'f4a78a6faacb4d8e97829f9888ac6740'
    >>> E = SyncableDict(key='f4a78a6faacb4d8e97829f9888ac6740')
    >>> E
    {'one': 1, 'two': 2}
"""
from __future__ import division, print_function, unicode_literals

import collections

from .dicts import Counter, DefaultDict, Dict
from .lists import Deque, List
from .sets import Set


class _SyncableBase(object):
    @property
    def redis(self):
        return self.persistence.redis

    @property
    def key(self):
        return self.persistence.key

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.sync()


class SyncableDict(_SyncableBase, dict):
    """
    :class:`dict` subclass whose contents can be synced to Redis.

    See Python's `dict documentation
    <https://docs.python.org/3/library/stdtypes.html#mapping-types-dict>`_ for
    details.
    """

    def __init__(self, **kwargs):
        self.persistence = Dict(**kwargs)

        super(SyncableDict, self).__init__()
        self.update(self.persistence)

    def sync(self):
        self.persistence.clear()
        self.persistence.update(self)


class SyncableCounter(_SyncableBase, collections.Counter):
    """
    :class:`collections.Counter` subclass whose contents can be synced to
    Redis.

    See Python's `Counter documentation
    <https://docs.python.org/3/library/collections.html#collections.Counter>`_
    for details.
    """

    def __init__(self, **kwargs):
        self.persistence = Counter(**kwargs)

        super(SyncableCounter, self).__init__()
        self.update(self.persistence)

    def sync(self):
        self.persistence.clear()
        self.persistence.update(self)


class SyncableDefaultDict(_SyncableBase, collections.defaultdict):
    """
    :class:`collections.defaultdict` subclass whose contents can be synced to
    Redis.

    See Python's `defaultdict documentation
    <https://docs.python.org/3/library/collections.html
    #collections.defaultdict>`_ for details.
    """

    def __init__(self, *args, **kwargs):
        self.persistence = DefaultDict(*args, **kwargs)

        super(SyncableDefaultDict, self).__init__(args[0] if args else None)
        self.update(self.persistence)

    def sync(self):
        self.persistence.clear()
        self.persistence.update(self)


class SyncableList(_SyncableBase, list):
    """
    :class:`list` subclass whose contents can be synced to Redis.

    See Python's `list documentation
    <https://docs.python.org/3/library/stdtypes.html
    #sequence-types-list-tuple-range>`_ for details.
    """

    def __init__(self, **kwargs):
        self.persistence = List(**kwargs)

        super(SyncableList, self).__init__()
        self.extend(self.persistence)

    def sync(self):
        self.persistence.clear()
        self.persistence.extend(self)


class SyncableDeque(_SyncableBase, collections.deque):
    """
    :class:`deque` subclass whose contents can be synced to Redis.

    See Python's `deque documentation
    <https://docs.python.org/3/library/collections.html#collections.deque>`_
    for details.
    """

    def __init__(self, iterable=None, maxlen=None, **kwargs):
        self.persistence = Deque(iterable=iterable, maxlen=maxlen, **kwargs)

        super(SyncableDeque, self).__init__(maxlen=self.persistence.maxlen)
        self.extend(self.persistence)

    def sync(self):
        self.persistence.clear()
        self.persistence.extend(self)


class SyncableSet(_SyncableBase, set):
    """
    :class:`set` subclass whose contents can be synced to Redis.

    See Python's `set documentation
    <https://docs.python.org/3/library/stdtypes.html
    #set-types-set-frozenset>`_ for details.
    """

    def __init__(self, **kwargs):
        self.persistence = Set(**kwargs)

        super(SyncableSet, self).__init__()
        self.update(self.persistence)

    def sync(self):
        self.persistence.clear()
        self.persistence.update(self)


class LRUDict(_SyncableBase, collections.MutableMapping):
    """
    :class:`dict`-like class that stores recently-used items in memory
    and automatically pushes older items to Redis.
    Useful for constraining a Python process's memory footprint when a Redis
    server with lots of memory is available.

    >>> D = LRUDict(maxsize=2)
    >>> D['a'] = 1  # 'a' is stored locally
    >>> D['b'] = 2  # 'b' is stored locally
    >>> D['c'] = 2  # 'a' is pushed to Redis and 'c' is stored locally
    >>> assert D['a'] == 1  # 'a' is retrieved from Redis and replaces 'b'
    >>> assert D['b'] == 2  # 'b' is retrieved from Redis and replaces 'c'
    >>> D['b'] = -2
    >>> D.sync()  # All items are copied to Redis
    """

    def __init__(self, maxsize=None, **kwargs):
        """
        :param maxsize: The number of items to keep in memory.
        :type maxsize: integer
        :param redis: Redis client instance. If not provided, default Redis
                      connection is used.
        :type redis: :class:`redis.StrictRedis`
        :param key: Redis key for the collection. Collections with the same key
                    point to the same data. If not provided, a random
                    string is generated.
        :type key: str
        """
        self.maxsize = None if (maxsize is None) else int(maxsize)
        self.cache = collections.OrderedDict()
        self.persistence = Dict(**kwargs)

        super(LRUDict, self).__init__()
        self.update(self.persistence)

    def __contains__(self, key):
        return (key in self.cache) or (key in self.persistence)

    def __delitem__(self, key):
        # Remove the given key from the local cache and the Redis cache
        for D in (self.cache, self.persistence):
            try:
                del D[key]
            except KeyError:
                pass

    def __getitem__(self, key):
        # If the item is in the local cache, grab its value and re-insert it
        # as the rightmost (i.e. most recently used) item.
        # Otherwise, try to get out of Redis and then insert it as the
        # rightmost item.
        try:
            value = self.cache.pop(key)
        except KeyError:
            value = self.persistence.pop(key)

        self[key] = value
        return value

    def __iter__(self):
        local_keys = set(self.cache.keys())
        persistence_keys = set(self.persistence.keys())

        return iter(local_keys | persistence_keys)

    def __len__(self):
        return sum(1 for k in self)

    def __setitem__(self, key, value):
        # If the key is in the local cache, remove it and re-insert it with
        # the new value as the rightmost (i.e. most recently used) item.
        # Otherwise, we'll add it to the local cache. If the local cache
        # is full we'll remove the leftmost (i.e. least recently used) item
        # and insert the new values as the rightmost item.
        try:
            self.cache.pop(key)
        except KeyError:
            if (self.maxsize is not None) and len(self.cache) >= self.maxsize:
                old_key, old_value = self.cache.popitem(last=False)
                self.persistence[old_key] = old_value

        self.cache[key] = value

    def clear(self):
        """
        Removes all items from both the local cache and the collection's Redis
        key.
        """
        self.cache.clear()
        self.persistence.clear()

    def copy(self, key=None):
        """
        Creates another collection with the same items and maxsize with
        the given *key*.
        """
        other = self.__class__(
            maxsize=self.maxsize, redis=self.persistence.redis, key=key
        )
        other.update(self)

        return other

    @classmethod
    def fromkeys(cls, seq, value=None, **kwargs):
        """
        Create a new collection with keys from *seq* and values set to
        *value*. The keyword arguments are passed to the persistent ``Dict``.
        """
        other = cls(**kwargs)
        other.update(((key, value) for key in seq))

        return other

    def sync(self, clear_cache=False):
        """
        Copy items from the local cache to the persistent Dict.
        If *clear_cache* is ``True``, clear out the local cache after
        pushing its items to Redis.
        """
        self.persistence.update(self)

        if clear_cache:
            self.cache.clear()



# -*- coding: utf-8 -*-
"""
sortedsets
~~~~~~~~~~

The `sortedsets` module contains a collection, :class:`SortedSetCounter`,
which provides an interface to Redis's
`Sorted Set <http://redis.io/commands#set>`_ type.
"""
from __future__ import division, print_function, unicode_literals

from .base import RedisCollection


class SortedSetCounter(RedisCollection):
    """
    :class:`SortedSetCounter` is a collection based on the Redis
    `Sorted Set <http://redis.io/topics/data-types#sorted-sets>`_ type.
    Instances map a unique set of ``member`` objects to floating point
    ``score`` values.

        >>> ssc = SortedSetCounter([('earth', 300), ('mercury', 100)])
        >>> ssc.set_score('venus', 200)
        >>> ssc.get_score('venus')
        200.0

    When retrieving members they are returned in order by score:

        >>> ssc.items()
        [('mercury', 100.0), ('venus', 200.0), ('earth', 300.0)]

    Ranges of items by rank can be computed and returned efficiently, as can
    ranges by score:

        >>> ssc.items(min_rank=1)  # 'mercury' has rank 0
        [('venus', 200.0), ('earth', 300.0)]
        >>> ssc.items(min_score=99, max_score=299)
        [('mercury', 100.0), ('venus', 200.0)]

    .. warning::
        The API for :class:`SortedSetCounter` does not attempt to match an
        existing Python collection's.

        - Unlike :class:`Dict` or :class:`Set` objects, equal numeric types are
          considered distinct when used as members. For example, a collection
          can contain both ``1`` and ``1.0``.

        - Unlike :class:`Counter` or :class:`collections.Counter` objects, only
          :class:`float` scores can be stored.
    """

    def __init__(self, *args, **kwargs):
        """
        Create a new SortedSetCounter object.

        If the first argument (*data*) is an iterable object, create the new
        SortedSetCounter with its elements as the initial data.

        :param data: Initial data.
        :type data: iterable or mapping
        :param redis: Redis client instance. If not provided, default Redis
                      connection is used.
        :type redis: :class:`redis.StrictRedis`
        :param key: Redis key for the collection. Collections with the same key
                    point to the same data. If not provided, a random
                    string is generated.
        :type key: str
        """
        data = args[0] if args else kwargs.pop('data', None)

        super(SortedSetCounter, self).__init__(**kwargs)

        if data:
            self.update(data)

    def _data(self, pipe=None):
        pipe = self.redis if pipe is None else pipe
        items = pipe.zrange(self.key, 0, -1, withscores=True)

        return [(self._unpickle(member), score) for member, score in items]

    def _repr_data(self):
        items = ('{}: {}'.format(repr(k), repr(v)) for k, v in self.items())
        return '{{{}}}'.format(', '.join(items))

    # Magic methods

    def __contains__(self, member, pipe=None):
        """Return ``True`` if *member* is present, else ``False``."""
        pipe = self.redis if pipe is None else pipe
        score = pipe.zscore(self.key, self._pickle(member))

        return score is not None

    def __iter__(self, pipe=None):
        """
        Return an iterator of ``(member, score)`` tuples from the collection.
        """
        pipe = self.redis if pipe is None else pipe

        return iter(self._data(pipe))

    def __len__(self, pipe=None):
        """Return the number of members in the collection."""
        pipe = self.redis if pipe is None else pipe

        return pipe.zcard(self.key)

    # Named methods

    def clear(self, pipe=None):
        self._clear(pipe=pipe)

    def copy(self, key=None):
        other = self.__class__(redis=self.redis, key=key)
        other.update(self)

        return other

    def count_between(self, min_score=None, max_score=None):
        """
        Returns the number of members whose score is between *min_score* and
        *max_score* (inclusive).
        """
        min_score = float('-inf') if min_score is None else float(min_score)
        max_score = float('inf') if max_score is None else float(max_score)

        return self.redis.zcount(self.key, min_score, max_score)

    def discard_by_rank(self, min_rank=None, max_rank=None, pipe=None):
        pipe = self.redis if pipe is None else pipe

        min_rank = 0 if min_rank is None else min_rank
        max_rank = -1 if max_rank is None else max_rank

        pipe.zremrangebyrank(self.key, min_rank, max_rank)

    def discard_by_score(self, min_score=None, max_score=None, pipe=None):
        pipe = self.redis if pipe is None else pipe

        min_score = float('-inf') if min_score is None else float(min_score)
        max_score = float('inf') if max_score is None else float(max_score)

        pipe.zremrangebyscore(self.key, min_score, max_score)

    def discard_between(
        self,
        min_rank=None,
        max_rank=None,
        min_score=None,
        max_score=None,
    ):
        """
        Remove members whose ranking is between *min_rank* and *max_rank*
        OR whose score is between *min_score* and *max_score* (both ranges
        inclusive). If no bounds are specified, no members will be removed.
        """
        no_ranks = (min_rank is None) and (max_rank is None)
        no_scores = (min_score is None) and (max_score is None)

        # Default scope: nothing
        if no_ranks and no_scores:
            return

        # Scope widens to given score range
        if no_ranks and (not no_scores):
            return self.discard_by_score(min_score, max_score)

        # Scope widens to given rank range
        if (not no_ranks) and no_scores:
            return self.discard_by_rank(min_rank, max_rank)

        # Scope widens to score range and then rank range
        with self.redis.pipeline() as pipe:
            self.discard_by_score(min_score, max_score, pipe)
            self.discard_by_rank(min_rank, max_rank, pipe)
            pipe.execute()

    def discard_member(self, member, pipe=None):
        """
        Remove *member* from the collection, unconditionally.
        """
        pipe = self.redis if pipe is None else pipe
        pipe.zrem(self.key, self._pickle(member))

    def get_score(self, member, default=None, pipe=None):
        """
        Return the score of *member*, or *default* if it is not in the
        collection.
        """
        pipe = self.redis if pipe is None else pipe
        score = pipe.zscore(self.key, self._pickle(member))

        if (score is None) and (default is not None):
            score = float(default)

        return score

    def get_or_set_score(self, member, default=0):
        """
        If *member* is in the collection, return its value. If not, store it
        with a score of *default* and return *default*. *default* defaults to
        0.
        """
        default = float(default)

        def get_or_set_score_trans(pipe):
            pickled_member = self._pickle(member)
            score = pipe.zscore(self.key, pickled_member)

            if score is None:
                pipe.zadd(self.key, default, self._pickle(member))
                return default

            return score

        return self._transaction(get_or_set_score_trans)

    def get_rank(self, member, reverse=False, pipe=None):
        """
        Return the rank of *member* in the collection.
        By default, the member with the lowest score has rank 0.
        If *reverse* is ``True``, the member with the highest score has rank 0.
        """
        pipe = self.redis if pipe is None else pipe
        method = getattr(pipe, 'zrevrank' if reverse else 'zrank')
        rank = method(self.key, self._pickle(member))

        return rank

    def increment_score(self, member, amount=1):
        """
        Adjust the score of *member* by *amount*. If *member* is not in the
        collection it will be stored with a score of *amount*.
        """
        return self.redis.zincrby(
            self.key, self._pickle(member), float(amount)
        )

    def items_by_rank(
        self, min_rank=None, max_rank=None, reverse=False, pipe=None
    ):
        pipe = self.redis if pipe is None else pipe

        min_rank = 0 if min_rank is None else min_rank
        max_rank = -1 if max_rank is None else max_rank

        if reverse:
            results = pipe.zrevrange(
                self.key, min_rank, max_rank, withscores=True
            )
        else:
            results = pipe.zrange(
                self.key, min_rank, max_rank, withscores=True
            )

        return [(self._unpickle(member), score) for member, score in results]

    def items_by_score(
        self, min_score=None, max_score=None, reverse=False, pipe=None
    ):
        pipe = self.redis if pipe is None else pipe

        min_score = float('-inf') if min_score is None else float(min_score)
        max_score = float('inf') if max_score is None else float(max_score)

        if reverse:
            results = pipe.zrevrangebyscore(
                self.key, max_score, min_score, withscores=True
            )
        else:
            results = pipe.zrangebyscore(
                self.key, min_score, max_score, withscores=True
            )

        return [(self._unpickle(member), score) for member, score in results]

    def items(
        self,
        min_rank=None,
        max_rank=None,
        min_score=None,
        max_score=None,
        reverse=False,
        pipe=None,
    ):
        """
        Return a list of ``(member, score)`` tuples whose ranking is between
        *min_rank* and *max_rank* AND whose score is between *min_score* and
        *max_score* (both ranges inclusive). If no bounds are specified, all
        items will be returned.
        """
        pipe = self.redis if pipe is None else pipe

        no_ranks = (min_rank is None) and (max_rank is None)
        no_scores = (min_score is None) and (max_score is None)

        # Default scope: everything
        if no_ranks and no_scores:
            ret = self.items_by_score(min_score, max_score, reverse, pipe)
        # Scope narrows to given score range
        elif no_ranks and (not no_scores):
            ret = self.items_by_score(min_score, max_score, reverse, pipe)
        # Scope narrows to given rank range
        elif (not no_ranks) and no_scores:
            ret = self.items_by_rank(min_rank, max_rank, reverse, pipe)
        # Scope narrows twice - once by rank and once by score
        else:
            results = self.items_by_rank(min_rank, max_rank, reverse, pipe)
            ret = []
            for member, score in results:
                if (min_score is not None) and (score < min_score):
                    continue
                if (max_score is not None) and (score > max_score):
                    continue
                ret.append((member, score))

        return ret

    def scan_items(self):
        """
        Yield each of the ``(member, score)`` tuples from the collection,
        without pulling them all into memory.

        .. warning::
            This method may return the same (member, score) tuple multiple
            times.
            See the `Redis SCAN documentation
            <http://redis.io/commands/scan#scan-guarantees>`_ for details.
        """
        for m, s in self.redis.zscan_iter(self.key):
            yield self._unpickle(m), s

    def set_score(self, member, score, pipe=None):
        """
        Set the score of *member* to *score*.
        """
        pipe = self.redis if pipe is None else pipe
        pipe.zadd(self.key, float(score), self._pickle(member))

    def update(self, other):
        """
        Update the collection with items from *other*. Accepts other
        :class:`SortedSetCounter` instances, dictionaries mapping members to
        numeric scores, or sequences of ``(member, score)`` tuples.
        """
        def update_trans(pipe):
            other_items = method(pipe=pipe) if use_redis else method()

            pipe.multi()
            for member, score in other_items:
                pipe.zadd(self.key, float(score), self._pickle(member))

        watches = []
        if self._same_redis(other, RedisCollection):
            use_redis = True
            watches.append(other.key)
        else:
            use_redis = False

        if hasattr(other, 'items'):
            method = other.items
        elif hasattr(other, '__iter__'):
            method = other.__iter__

        self._transaction(update_trans, *watches)
